#include "filters.h"

filetype_t filetype=FUNKNOWN;
options_t options={NULL,NULL,NULL,NULL,NULL,NULL,NULL,false,NULL,NULL,NULL};
int pagecount[2]={0,0};

err_t do_banner(void)
{
  char *argv[64];
  extern char **environ;

  dprintf("do_banner()\n");

  argv[0]="/fs/lprng/bin/psfbanner";
  argv[1]=options.username;
  argv[2]=options.realname;
  argv[3]=options.printername;
  argv[4]=options.comment;
  argv[5]="No Z opts";
  argv[6]=strdup(now());
  argv[7]=options.hostname;
  argv[8]=options.jobname;
  argv[9]=options.jobnumber;
  argv[10]=options.jobname;
  argv[11]=NULL;

  forkfilter(argv,environ);
}

char *printable(int c)
{
  static char *low[]={"NULL","^A","^B","^C","^D","^E","^F","^G","^H","^I",
		      "line-feed","^K","^L","carriage-return","^N","^O",
		      "^P","^Q","^R","^S","^T","^U","^V","^W","^X","^Y",
		      "^Z","ESC","^\\","^]","^^","^_"};
  static char buf[3];

  if (c=='\n') return("\n");
  if (c==' ') return(" ");
  if (c=='\t') return("\t");
  
  if (c<32)
    return(low[c]);

  if (c < 127)
    {
      sprintf(buf,"%c",c);
      return(buf);
    }

  if (c == 127)
    return("^?");

  if (c<256)
    {
      sprintf(buf,"%02x",c);
      return(buf);
    }

  return("internal-error");
}

err_t readoptions(int argc, char **argv)
{
  int i;
  struct passwd *pwd;

  for (i=1;i<argc;i++)
    {
      dprintf("argv[%d]: %s",i,argv[i]);
      if (strncmp("-P",argv[i],2)==0)
	options.printername=strdup(argv[i]+2);
      else if (strncmp("-n",argv[i],2)==0)
	options.username=strdup(argv[i]+2);
      else if (strncmp("-J",argv[i],2)==0)
	options.jobname=strdup(argv[i]+2);
      else if (strncmp("-j",argv[i],2)==0)
	options.jobnumber=strdup(argv[i]+2);
      else if (strncmp("-h",argv[i],2)==0)
	options.hostname=strdup(argv[i]+2);
      else if (strncmp("-Z",argv[i],2)==0)
	options.zopts=strdup(argv[i]+2);
      else if (strncmp("-e",argv[i],2)==0)
	options.datapath=strdup(argv[i]+2);
      else if (strncmp("--banner",argv[i],8)==0)
	options.banner=true;
      else if (strncmp("-S",argv[i],2)==0)
	options.comment=strdup(argv[i]+2);
      else if (strncmp("-Q",argv[i],2)==0)
	options.queuename=strdup(argv[i]+2);
      else 	
	dprintf(": UNUSED");
      dprintf("\n");
    }

  if (!options.printername)
    {
      dprintf("readoptions(): no printername\n");
      return(e_err);
    }
  if (!options.hostname)
    {
      dprintf("readoptions(): no hostname\n");
      return(e_err);
    }
  if (!options.jobname)
    {
      dprintf("readoptions(): no jobname\n");
      return(e_err);
    }
  if (!options.username)
    {
      dprintf("readoptions(): no username\n");
      return(e_err);
    }
  if (!options.datapath && !options.banner)
    {
      dprintf("readoptions(): no datapath\n");
      return(e_err);
    }
  if (!options.jobnumber)
    {
      dprintf("readoptions(): no jobnumber\n");
      return(e_err);
    }

  if (!options.comment)
    options.comment="This printer has no comment.";

  pwd=getpwnam(options.username);
  if (pwd)
    options.realname=strdup(pwd->pw_gecos);
  else
    {
      dprintf("readoptions(): user %s does not exist?\n",options.username);
      return(e_err);
    }

  return(e_ok);
}

err_t legacyacct(int s, int f)
{
  FILE *fp;
  char buf[512];
  
  fp=fopen(ACCTDIR ACCTFILE,"a");
  if (!fp)
    {
      perror("fopen(\"" ACCTDIR ACCTFILE "\",\"a\")");
      return(e_err);
    }

  fprintf(fp,"%s|%s|%s|%d|%s|%d|%s\n",
	      options.hostname,options.username,
	      (options.queuename?options.queuename:options.printername),
	      (f-s),now(),time(NULL),
	      (options.banner?"(banner)":options.jobname));
  fclose(fp);
  return(e_ok);
}

err_t doacct(char *ff, int pagecount)
{
  FILE *fp;
  char buf[512];

  sprintf(buf,"%s/%s",ACCTDIR,options.printername);
  fp=fopen(buf,"a+");
  if (!fp) perror("fopen()");

  if (fprintf(fp,"%s|%s|%s|%s|%d|%s|%ld|%s\n",
	      ff,options.hostname,options.username,options.printername,
	      pagecount,now(),time(NULL),
	      (options.banner?"(banner)":options.jobname))<0)
    {
      perror("fprintf()");
      fclose(fp);
      return(e_err);
    }
  if (fp != stderr)
    {
      fflush(fp);
      fclose(fp);
    }
  return(e_ok);
}

char *suckdata(int fd)
{
  char *ptr=NULL;
  int len=512;
  int bytes;

  int index=0;

  ptr=(char *)malloc(len);
  if (ptr==NULL)
    {
      perror("malloc()");
      return(NULL);
    }
  
  while((bytes=read(fd,ptr+index,len-index))>0)
    {
      index += bytes;
      if (index==len)
	{
	  ptr=(char *)realloc(ptr,len*2);
	  if (ptr==NULL)
	    {
	      perror("realloc()");
	      return(NULL);
	    }
	  len *= 2;
	}
    }
  if (bytes<0) perror("read()");

  return(ptr);
}

int openprinter(char *printer, unsigned short port)
{
  struct sockaddr_in srv_addr;
  struct hostent host_info;
  struct hostent *hp;
  unsigned long inaddr;
  int fd;
  int tries;

  bzero((char *)&srv_addr,sizeof(srv_addr));
  srv_addr.sin_family=AF_INET;
  srv_addr.sin_port=htons(port);

  inaddr=inet_addr(printer);
  if (inaddr!=0xffffffff)
    {
      bcopy((char *)&inaddr,(char *)&srv_addr.sin_addr,sizeof(inaddr));
      host_info.h_name=NULL;
    }
  else
    {
      hp=gethostbyname(printer);
      if (hp==NULL)
	{
	  perror("gethostbyname()");
	  return(-1);
	}
      host_info=*hp;
      bcopy(hp->h_addr,(char *)&srv_addr.sin_addr,hp->h_length);
    }

  if ((fd=socket(AF_INET,SOCK_STREAM,0))<0)
    {
      perror("socket()");
      return(-1);
    }

  while((connect(fd,(struct sockaddr *)&srv_addr,sizeof(srv_addr)))<0)
    {
      if (close(fd)<0) perror("close()");
      perror("connect()");
      sleep(1);
      tries++;
      if (tries==100)
	{
	  perror("connect()");
	  return(-1);
	}
      else
	{
	  dprintf("retrying...\n");
	  if ((fd=socket(AF_INET,SOCK_STREAM,0))<0)
	    {
	      perror("socket()");
	      return(-1);
	    }
	}
    }

  return(fd);
}

int getpagecount(void)
{
  int fd;
  const char *pagecountps="(%%[ pagecount: ) print statusdict begin pagecount end 20 string cvs print ( ]%%\n) print flush";
  char *output;
  int pages=0;

  fd=openprinter(options.printername,9100);
  if (fd<0)
    {
      dprintf("failed to open port to the printer\n");
      return(0);
    }

  if (writebuf(fd,pagecountps,strlen(pagecountps))==e_err)
    perror("write()");

  if (shutdown(fd,1)<0)
    perror("shutdown");

  usleep(500); /* wait half a second */

  output=suckdata(fd);
  if (output)
    {
      sscanf(output,"%%%%[ pagecount: %d ]%%%%",&pages);
      free(output);
    }

  if (close(fd)<0) perror("close()");
  usleep(500); /* give the phaser a little time actually close() */

  return(pages);
}

void ouch(int sig)
{
  switch(sig)
    {
    case SIGPIPE:
      return;
      break;
    case SIGINT:
      dprintf("got a SIGINT\n");
      exit(JABORT);
      break;
    default:
      dprintf("got signal: %d\n",sig);
      fflush(stderr);
      kill(getpid(),sig);
      break;
    }
  dprintf("WHEN WILL THE HURTING STOP?\n");
  fflush(stderr);
  exit(JABORT);
}

void setupsignals(void)
{
  int i;

  for (i=1;i<NSIG;i++);
  signal(i,&ouch);
}

err_t init(int argc, char **argv)
{
  char buf[512];

  dprintf("calling setupsignals()\n");
  setupsignals();
  dprintf("called setupsignals()\n");

  dprintf("calling readoptions()\n");
  if (readoptions(argc, argv)==e_err)
    {
      return(e_err);
    }
  dprintf("called readoptions()\n");

  sprintf(buf,"%s %s %s",SENDMAIL,PADMIN,options.username);

  dprintf("taking care of a banner page\n");
  if (options.banner)
    {
      do_banner();
      return(e_ok);
    }
  dprintf("took care of a banner page\n");

  /*
  ** what kind of file are we printing?
  */
  dprintf("checking file type\n");
  filetype=getfiletype(options.datapath);
  if (filetype==FERROR)
    {
      dprintf("Could not determine file type!\n");
    }
  else
    {
      return(setupfilters(filetype));
    }
  dprintf("I can't print this.\n");
  exit(JREMOVE);
}

int main(int argc, char **argv, char **envp)
{
  int fd;
  char buf[CHUNK];
  int bytes;
  int i;

  dprintf("-----------------------------------------------------------------------------\n");
  dprintf("calling init()\n");
  if (init(argc, argv)==e_err)
    {
      exit(JABORT);
    }
  dprintf("called init()\n");

  dprintf("calling getpagecount()\n");
  while (pagecount[0]==0)
    {
      pagecount[0]=getpagecount();
    }
  dprintf("called getpagecount()\n");
  /*
  **  doacct("START",pagecount[0]);
  */

  dprintf("calling openprinter()\n");
  fd=openprinter(options.printername,9100);
  if (fd<0) exit(JABORT);
  dprintf("called openprinter()\n");

  while((bytes=read(0,buf,CHUNK))>0)
    if (writebuf(fd,buf,bytes)==e_err)
      perror("write()");

  if (shutdown(fd,1)<0)
    perror("shutdown");

  while ((bytes=read(fd,buf,CHUNK))!=0)
    {
      for (i=0;i<bytes;i++)
	{
	  dprintf("%s",printable(buf[i]));
	}
    }
  dprintf("\n");

  if (close(fd)<0) perror("close()");

  while (pagecount[1] == 0)
    {
      pagecount[1]=getpagecount();
    }

  /*
  **  doacct("END",pagecount[1]);
  */

  legacyacct(pagecount[0],pagecount[1]);

  return(JSUCC);
}
